/**
 * WEBLAB: Service oriented integration platform for media mining and intelligence applications
 * 
 * Copyright (C) 2004 - 2011 Cassidian, an EADS company
 * 
 * This library is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License along with this
 * library; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA 02110-1301 USA
 * 
 * /!\ At the time of the 1.2 release, this module is still experimental and need more in 
 * depth use and validation. Use it at your own risks ;-)
 */

package org.ow2.weblab.rdf;

import java.net.URI;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.transform.dom.DOMResult;

import org.apache.commons.logging.LogFactory;
import org.ow2.weblab.core.annotator.Element;
import org.ow2.weblab.core.model.PieceOfKnowledge;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

public final class RDFWriter {

	private static XMLOutputFactory outputFactory = null;
	protected static final String XML_PARSER_CLASS = "com.sun.xml.internal.stream.XMLOutputFactoryImpl";

	private RDFWriter() {
		throw new UnsupportedOperationException("This class only contains " + "static methods; no need to instantiate it.");
	}

	public static void writeStatement(PieceOfKnowledge annotation, URI subject, String prefix, URI namespace, String property, Object value, String lang,
			boolean add, Element element) throws XMLStreamException, ParserConfigurationException {
		
		if (value == null && add){
			throw new XMLStreamException("You can not delete a statement using a null value when the annotator is in append mode.");
		}
		
		Object o = annotation.getData();
		if (o == null) {
			if (value == null){
				throw new XMLStreamException("You tried to delete a statement ("+subject+", "+namespace+property+", "+value+") that does not exist.");
			}
			createAnnotation(annotation, subject, prefix, namespace.toString(), property, value, lang, element);
		} else {
			updateAnnotation(annotation, subject, prefix, namespace.toString(), property, value, lang, add, element);
		}
	}
	
	private static void updateAnnotation(PieceOfKnowledge annotation, URI subject, String prefix, String namespace, String property, Object value, String lang,
			boolean add, Element element) throws XMLStreamException {
		Node n = (Node) annotation.getData();

		// We search for an rdf description about the subject
		Node description = RDFReader.findElementAbout(n, subject.toString(), element);

		if (description == null) {
			// we create a new description node
			description = createElementAbout(RDFReader.findRDF(n), subject.toString(), element);
			if (description == null) {
				throw new XMLStreamException("Can not create rdf:description about : " + subject);
			}
		} else if (!add) {
			List<Node> nodes = RDFReader.findNodesWithProperty(description, namespace, property);

			for (Node node : nodes) {
				if (lang != null) {
					String nlang = RDFReader.getLanguage(node);
					if (lang.equalsIgnoreCase(nlang)) {
						node.getParentNode().removeChild(node);
					}
				} else {
					node.getParentNode().removeChild(node);
				}
			}

			// if the value is null
			// we delete the statment
			if (value == null){
				return;
			}
		}

		DOMResult domResult = new DOMResult(description);

		XMLStreamWriter xmlw = createXMLStreamWriter(domResult);
		writeAnnotation(xmlw, prefix, namespace, property, value, lang);
		xmlw.close();
		annotation.setData(n);

	}
	
	/**
	 * Remove all or part of a statement
	 * TODO: add check for null values
	 * @param annotation
	 * @param subject
	 * @param prefix
	 * @param namespace
	 * @param property
	 * @param value
	 * @param lang
	 * @param add
	 * @param element
	 */
	public static void removeStatement(PieceOfKnowledge annotation, URI subject, String prefix, URI namespace, String property, Object value, String lang,
			boolean add, Element element){
		Node n = (Node) annotation.getData();
		Node description = RDFReader.findElementAbout(n, subject.toString(), element);
		
		if (description != null){
			if (property == null){
				description.getParentNode().removeChild(description);
				return;
			}
			List<Node> nodes = RDFReader.findNodesWithProperty(description, namespace.toString(), property);
			for (Node node : nodes) {
				String nodeValue = RDFReader.getNodeLiteralOrResource(node);
				if (lang != null) {
					String nlang = RDFReader.getLanguage(node);
					if (lang.equalsIgnoreCase(nlang) && nodeValue.equals(value.toString())) {
						
						node.getParentNode().removeChild(node);
					}
				} else if(nodeValue.equals(value.toString())){
					node.getParentNode().removeChild(node);
				}
			}
		}
	}

	private static XMLStreamWriter createXMLStreamWriter(DOMResult domResult) throws XMLStreamException {
		if (outputFactory == null) {
			try {
				/**
				 * Try to prevent XML API overriding.
				 * Sun XML parser is valid with namespace and xml:lang support in XML.
				 */
				Class<?> clazz = ClassLoader.getSystemClassLoader().loadClass(XML_PARSER_CLASS);
				outputFactory = (XMLOutputFactory) clazz.newInstance();
			} catch (ClassNotFoundException e) {
				LogFactory.getLog(RDFWriter.class).warn(e);
			} catch (InstantiationException e) {
				LogFactory.getLog(RDFWriter.class).warn(e);
			} catch (IllegalAccessException e) {
				LogFactory.getLog(RDFWriter.class).warn(e);
			}
			if (outputFactory == null) {
				LogFactory.getLog(RDFWriter.class).warn(
						"Your JVM is not the Sun one, this is not really a problem " + "(we like open source too) but please define an implementation "
								+ "of a XMLOutputFactory in META-INF/services/javax.xml.stream.XMLOutputFactory");
				/**
				 * Your JVM is not the Sun one, this is not really a problem (we like open source too) but
				 * please define an implementation of a XMLOutputFactory in META-INF/services/javax.xml.stream.XMLOutputFactory
				 * See http://java.sun.com/javase/6/docs/api/javax/xml/stream/XMLOutputFactory.html#newFactory() for more information
				 */
				outputFactory = XMLOutputFactory.newInstance();
			}
		}

		return outputFactory.createXMLStreamWriter(domResult);
	}

	private static Node createElementAbout(Node rdf, String subject, Element element) throws XMLStreamException {
		DOMResult domResult = new DOMResult(rdf);

		XMLStreamWriter xmlw = createXMLStreamWriter(domResult);

		xmlw.writeStartElement(element.getPrefix(), element.getLocalName(), element.getUri());
		xmlw.writeAttribute(CommonNamespaces.RDF_PREFIX, CommonNamespaces.RDF_URI, "about", subject);
		xmlw.writeEndElement();
		xmlw.close();

		return RDFReader.findElementAbout(rdf, subject, element);
	}

	private static void createAnnotation(PieceOfKnowledge annotation, URI subject, String prefix, String namespace, String property, Object value, String lang,
			Element element) throws XMLStreamException, ParserConfigurationException {

		DocumentBuilderFactory documentBuilderFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = documentBuilderFactory.newDocumentBuilder();
		Document document = builder.newDocument();
		DOMResult domResult = new DOMResult(document);

		XMLStreamWriter xmlw = createXMLStreamWriter(domResult);

		// Add prefixes
		xmlw.writeStartElement("data");
		xmlw.writeStartElement(CommonNamespaces.RDF_PREFIX, "RDF", CommonNamespaces.RDF_URI);
		xmlw.writeNamespace(CommonNamespaces.RDF_PREFIX, CommonNamespaces.RDF_URI);
		xmlw.writeNamespace(prefix, namespace);

		xmlw.writeStartElement(element.getPrefix(), element.getLocalName(), element.getUri());
		xmlw.writeNamespace(prefix, namespace);
		xmlw.writeAttribute(CommonNamespaces.RDF_PREFIX, CommonNamespaces.RDF_URI, "about", subject.toString());

		// Add annotation
		writeAnnotation(xmlw, prefix, namespace, property, value, lang);

		xmlw.writeEndElement();

		xmlw.writeCharacters("\n");
		xmlw.writeCharacters("        ");
		xmlw.writeEndElement();
		xmlw.writeEndElement();
		xmlw.writeEndDocument();
		xmlw.close();

		annotation.setData(document.getDocumentElement());

	}

	private static void writeAnnotation(XMLStreamWriter xmlw, String prefix, String namespace, String property, Object value, String lang)
			throws XMLStreamException {

		// Check of properties
		if (property.indexOf(':') != -1 && property.indexOf('/') != -1) {
			int l = property.lastIndexOf('/');
			int d = property.lastIndexOf('#');
			d = d == -1 ? l : d;
			namespace = property.substring(0, d + 1);
			prefix = "w" + namespace.hashCode();
			property = property.substring(d + 1);
		}

		xmlw.writeNamespace(prefix, namespace);
		if (value instanceof URI) {
			writeResource(xmlw, prefix, namespace, property, value.toString(), lang);
		
//		}else if(value == null){
//			TODO:  delete the annotation
			
		}else{
			writeLiteral(xmlw, prefix, namespace, property, value.toString(), lang);
		}
	}

	private static void writeResource(XMLStreamWriter xmlw, String prefix, String namespace, String property, String value, String lang)
			throws XMLStreamException {
		xmlw.writeStartElement(prefix, property, namespace);
		xmlw.writeAttribute(CommonNamespaces.RDF_PREFIX, CommonNamespaces.RDF_URI, "resource", value);
		if (lang != null) {
			xmlw.writeAttribute("xml:lang", lang);
		}
		xmlw.writeEndElement();
	}

	private static void writeLiteral(XMLStreamWriter xmlw, String prefix, String namespace, String property, String value, String lang)
			throws XMLStreamException {
		xmlw.writeStartElement(prefix, property, namespace);
		if (lang != null) {
			xmlw.writeAttribute("xml:lang", lang);
		}
		xmlw.writeCharacters(value);
		xmlw.writeEndElement();
	}
	
	/**
	 * Replace all uris in a pok by another 
	 * @param pok
	 * @param source
	 * @param dest
	 * @return
	 */
	public static PieceOfKnowledge replaceURI(PieceOfKnowledge pok, URI source, URI dest){
		Node n = (Node) pok.getData();
		// find about element
		Node about = RDFReader.findElementAbout(n,source.toString(),Element.getDefaultElement());
		if (about != null){
			Node abn = about.getAttributes().getNamedItemNS(CommonNamespaces.RDF_URI, "about");
			abn.setNodeValue(dest.toString());
		}
		
		// find resources
//		List<Node> nodes = RDFReader.findNodesWithValue(n, source.toString());
//		for(Node node:nodes){
//			System.out.println("change resource for ! "+node);
//		}
		
		return pok;
	}

}
